local M = {}

local scenario = require("scenario/scenarios")
local helper = require("scenario/scenariohelper")
local stats = require("statistics/statistics")

local waypointPrefix = "or_wp"
local racerPrefix = "or_racer"
local player, racers, waypoints, places, names, score = {}, {}, {}, {}, {}, {}
local racerNames = {"Gull", "Arnok", "Bob", "David", "Neil", "YBR", "CJ", "Niko", "Razor", "Mia", "Darius", "Sonic", "Amy", "Mario", "Luigi", "Shingo", "Itsuki", "Snake", "Xenon", "Danny", "Erin", "Gene", "Harry", "Sam", "Pedro"}
local text = {pass = "is eliminated!", spin = "is mad at you!", crash = "is smashed up!", bigCrash = "is obliterated!", win = "Congratulations, you made it!", fail = "Ouch, too bad! Try again!"}
local winText = {pass = "racers got eliminated.", crash = "racers got wrecked.", bonus = "Wow, you sure know how to create an entertaining show!"}
local mainMap = nil
local win = false
local queue = 0
local wpSplit = 2
local ahead = 3
local parTime = 220
local limitScore = 1000
local totalScore = 3000
local crawlSpeed = 3
local badDamage = 10000
local nearDist = 40
local farDist = 500
local actionDist = 100

local function setSpawn(v) -- teleports a vehicle to a preset spawn point
	local n = player.zone == 0 and v.zone or player.zone + ahead
	local pos = places[n]
	local rot = vec3(0, 0, 0)
	local moveUp = vec3(0, 0, 0.3)
	local moveSide = vec3(0, 0, 0)
	local vUp = vec3(0, 0, 1)
	local allow = true
	if places[n + 1] then
		rot = (places[n + 1] - places[n]):normalized()
		moveSide = rot:z0():cross(vUp)
	else
		rot = vec3(player.obj:getDirectionVector())
	end
	for i, w in ipairs(racers) do
		if player.zone > 0 and v.obj.name ~= w.obj.name then
			if ((places[n] + rot * 15) - w.pos):length() <= nearDist then
				allow = false
			end
		end
	end
	if allow then
		pos = pos + moveUp + moveSide
		rot = quatFromDir(vec3(0, -1, 0):rotated(quatFromDir(rot, vUp)), vUp)
		v.obj:setPositionRotation(pos.x, pos.y, pos.z, 0, 0, rot.z, rot.w)
		v.obj:queueLuaCommand("vehicleResetted(false)")
		v.obj:resetBrokenFlexMesh()
		local color = {}
		if player.zone > 0 then
			color = v.colors[math.random(1, #v.colors)]
			v.obj.color = ColorF(color[1], color[2], color[3], color[4]):asLinear4F()
		end
		v.zone = n
		v.state = "set"
	end
end

local function setPath(v) -- sets the race path for AI vehicles ahead of the player vehicle
	local newWaypoints = {}
	for i, w in ipairs(waypoints) do
		if i >= (v.zone / wpSplit) + 1 then
			table.insert(newWaypoints, w)
		end
	end
	local arg = {vehicleName = v.obj.name, waypoints = newWaypoints, aggression = v.aggro}
	helper.setAiPath(arg)
end

local function getRandomList(needed, total) -- returns a list of non-repeating random integers
	if needed > total then return nil end
	local list1, list2 = {}, {}
	for i = 1, total, 1 do
		table.insert(list1, i)
	end
	for i = 1, needed, 1 do
		local key = math.random(1, #list1)
		table.insert(list2, list1[key])
		table.remove(list1, key)
	end
	return list2
end

local function createSpawnPoints(total, split) -- creates spawn points at waypoints and road nodes between waypoints
	if not split then split = 1 end
	for i = 1, total, 1 do
		local j = i * split - (split - 1)
		local pos1 = vec3(scenetree.findObject(waypoints[i]):getPosition())
		places[j] = pos1
		if waypoints[i + 1] then
			local n = 1
			local pos2 = vec3(scenetree.findObject(waypoints[i + 1]):getPosition())
			while j % split ~= 0 do
				local posTemp = linePointFromXnorm(pos1, pos2, n / split)
				local posMid = mainMap.nodes[map.findClosestRoad(posTemp)].pos
				j, n = j + 1, n + 1
				places[j] = posMid
				if n > 10 then break end
			end
		end
	end
end

local function findNearPlace() -- finds the nearest place to the player, if close enough
	local n, place = player.zone, nil
	for i, v in ipairs(places) do
		if math.abs(player.zone - i) <= 3 then 
			local tempDist = (player.pos - places[i]):length()
			if tempDist <= nearDist / 2 then
				n = i
				place = places[n]
			end
		end
	end
	return n, place
end

local function getColors(jbeam) -- returns a vehicle's paint color choices
	local baseColors, colors = {}, {}
	local json = jsonReadFile and jsonReadFile("/vehicles/"..jbeam.."/info.json") or nil
	baseColors = json and json["colors"] or {white = "1 1 1 1.4", black = "0 0 0 1.4", red = "1 0 0 1.4", blue = "0 0 1 1.4"}
	for k, v in pairs(baseColors) do
		table.insert(colors, stringToTable(baseColors[k], " "))
	end
	return colors
end

local function onScenarioUIReady()
	player, racers, waypoints, places, names, score = {}, {}, {}, {}, {}, {}
	player.obj = be:getPlayerVehicle(0)
	player.failTimer = 0
	player.zone = 0
	win = false
	queue = 0
	score.pass, score.time, score.crash, score.realPass, score.realCrash = 0, 0, 0, 0, 0
	mainMap = map.getMap()
	local waypointsTotal = 0
	local waypointsAll = scenetree.findClassObjects("BeamNGWaypoint")
	for i, v in ipairs(waypointsAll) do
		if string.find(v, waypointPrefix) then
			waypointsTotal = waypointsTotal + 1
		end
	end
	for i = 1, waypointsTotal, 1 do
		table.insert(waypoints, waypointPrefix..i)
	end
	createSpawnPoints(waypointsTotal, wpSplit)
	local racersTotal = 0
	local racersAll = scenetree.findClassObjects("BeamNGVehicle")
	for i, v in ipairs(racersAll) do
		if string.find(v, racerPrefix) then
			racersTotal = racersTotal + 1
		end
	end
	local numList = getRandomList(#racerNames, #racerNames)
	for i, v in ipairs(numList) do
		table.insert(names, racerNames[v])
	end
	numList = getRandomList(racersTotal, racersTotal)
	for i = 1, racersTotal, 1 do
		local v = {}
		v.obj = scenetree.findObject(racerPrefix..numList[i])
		v.obj:queueLuaCommand("ai.setMode('random')")
		v.nickname = names[#names]
		table.remove(names)
		v.colors = getColors(v.obj.jbeam)
		v.state = "set"
		v.action = {}
		v.zone = i
		v.delay = true
		setSpawn(v)
		table.insert(racers, v)
	end
end

local function onRaceStart()
	stats.initialiseArbitraryStat("outrun", "Outrun", player.obj, player.obj.name, limitScore, limitScore)
	stats.initialiseArbitraryStat("timeScore", "Time", player.obj, player.obj.name, limitScore, limitScore)
	stats.initialiseArbitraryStat("aggro", "Aggression", player.obj, player.obj.name, limitScore, limitScore)
end

local function onRaceTick(tick)
	player.pos = vec3(player.obj:getPosition())
	player.vel = map.objects[map.objectNames[player.obj.name]].vel:length()
	player.vec = vec3(player.obj:getDirectionVector())
	player.damage = map.objects[map.objectNames[player.obj.name]].damage
	local newPlace = findNearPlace()
	player.zone = player.zone < newPlace and newPlace or player.zone
	if player.zone == #places then
		win = true
	end
	if not win and player.damage >= badDamage then
		if player.vel <= crawlSpeed then
			player.failTimer = player.failTimer + tick
			if player.failTimer >= math.max(2, (badDamage * 15) / player.damage) then -- player fails based on damage and ability to move
				scenario.finish({failed = text.fail})
			end
		else
			player.failTimer = 0
		end
	end
	for i, v in ipairs(racers) do
		v.pos = vec3(v.obj:getPosition())
		v.dist = (player.pos - v.pos):length()
		v.vel = map.objects[map.objectNames[v.obj.name]].vel:length()
		v.vec = vec3(v.obj:getDirectionVector())
		v.damage = map.objects[map.objectNames[v.obj.name]].damage
		if v.state == "set" then
			if v.delay then
				if player.zone > 0 then
					v.obj:queueLuaCommand("ai.setMode('random')")
				end
				v.nickname = names[#names]
				if #names > 1 then
					table.remove(names)
				end
				v.aggro = math.random(100, 130) / 100
				v.obj:queueLuaCommand("ai.setAggression("..tostring(v.aggro)..")")
				setPath(v)
				v.uptime = 0
				v.action = {}
				v.state = "normal"
				v.delay = false
			end
			v.delay = true
		end
		if (v.state == "normal" or v.state == "almost") and v.uptime >= 3 then
			if v.dist >= farDist then
				v.state = "ready"
			end
			if (player.pos - (v.pos + v.vec * 15)):length() <= nearDist then
				v.action.battle = true
			end
			if v.dist <= actionDist then
				if v.damage >= badDamage * 3 then
					if not v.action.crash then
						helper.flashUiMessage(v.nickname.." "..text.bigCrash, 2, false)
						score.crash = score.crash + 300
						score.realCrash = score.realCrash + 1
						v.obj:queueLuaCommand("ai.setMode('stopping')")
						v.obj:queueLuaCommand("fire.explodeVehicle()")
						v.action.crash = true
					end
				elseif v.damage >= badDamage then
					if not v.action.crash then
						helper.flashUiMessage(v.nickname.." "..text.crash, 2, false)
						score.crash = score.crash + 200
						score.realCrash = score.realCrash + 1
						v.obj:queueLuaCommand("ai.setMode('stopping')")
						v.action.crash = true
					end
				end
				if not v.action.spin and not v.action.crash then
					if v.dist <= 15 then
						if player.vec:dot(v.vec) < -0.8 and (player.vel >= 5 and v.vel >= 5) then -- rough detection of a PIT maneuver
							helper.flashUiMessage(v.nickname.." "..text.spin, 2, false)
							score.crash = score.crash + 100
							v.action.spin = true
						end
					end
				end
				if not v.action.pass and v.action.battle then
					if places[player.zone + 1] then
						if (places[player.zone + 1] - v.pos):length() - (places[player.zone + 1] - player.pos):length() >= nearDist then
							if (places[player.zone + 1] - player.pos):normalized():dot((player.pos - v.pos):normalized()) >= 0 then -- AI racer fell behind player
								helper.flashUiMessage(v.nickname.." "..text.pass, 2, false)
								score.pass = score.pass + 100
								score.realPass = score.realPass + 1
								v.obj:queueLuaCommand("ai.setMode('stopping')")
								v.action.pass = true
								v.action.crash = true
								v.state = "almost"
							end
						end
					end
				end
			end
			if v.state == "almost" then
				if v.dist >= actionDist then
					v.state = "ready"
				end
			end
			if player.zone > #places - ahead * 2 then
				if (places[#places] - v.pos):length() <= nearDist then
					v.obj:queueLuaCommand("ai.setMode('random')")
					v.state = "none"
				end
			end
		end
		if v.state == "ready" and player.zone <= #places - ahead then
			if queue == i then
				setSpawn(v)
				queue = -2
			end
		end
		v.uptime = v.uptime + tick
	end
	queue = queue >= #racers and 1 or queue + 1
end

local function onRaceResult()
	if win then
		local finishTime = scenario.getScenario().timer
		local finishText = text.win
		score.pass = math.min(limitScore, score.pass)
		score.crash = math.min(limitScore, score.crash)
		score.time = math.min(limitScore, math.floor(parTime + 100 - finishTime) * 10)
		stats.setStatProgress(player.obj:getID(), "outrun", player.obj.name, {value = score.pass})
		stats.setStatProgress(player.obj:getID(), "timeScore", player.obj.name, {value = score.time})
		stats.setStatProgress(player.obj:getID(), "aggro", player.obj.name, {value = score.crash})
		finishText = finishText.." "..score.realPass.." "..winText.pass.." "..score.realCrash.." "..winText.crash
		if (score.pass + score.time + score.crash) / totalScore >= 0.9 then
			finishText = finishText.." "..winText.bonus
		end
		scenario.finish({msg = finishText})
	end
end

M.onScenarioUIReady = onScenarioUIReady
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult

return M